require_relative 'config/environment_attribute_reader'

module BranchableCDNAssets
  class Config
    include EnvironmentAttributeReader
    extend  EnvironmentAttributeReader::ClassMethods

    attr_reader :raw_data, :branch,
                :production_branch, :cdn_dir, :file_filter,
                :dir_permissions, :file_permissions, :rsync_flags
    env_attr_reader :host, :root, :url, :allow_local

    def initialize data, branch=Asgit.current_branch
      @raw_data = normalize_data data
      @branch   = branch

      @production_branch = raw_data.fetch :production_branch, 'master'
      @cdn_dir           = raw_data.fetch :dir,               'cdn'
      @file_filter       = raw_data.fetch :file_filter,       nil
      @dir_permissions   = env_attr(:dir_permissions)      || 755
      @file_permissions  = env_attr(:file_permissions)     || 644
      @rsync_flags       = env_attr(:rsync_flags)          || '-aviz'
      @allow_local       = env_attr(:allow_local)          || false

      if @file_filter && ( !@file_filter.respond_to?(:arity) || @file_filter.arity != 1 )
        raise ArgumentError, "file_filter must be a lambda accepting one argument"
      end
    end

    def env
      @env ||= env_for_branch
    end

    def environments
      @environments ||= add_env_path_modifications_to_env_data raw_data.fetch(:environments, {})
    end

    def remotes
      return @_remotes if @_remotes

      @_remotes = environments[env.to_sym].fetch(:remotes, [default_remote]).map do |r|
        remote = Remote.new
        default_remote.merge( normalize_data(r) ).each do |k, v|
          remote.public_send "#{k}=", v
        end
        remote
      end
    end

    def invalidator
      return @_invalidator if @_invalidator

      invalidator_data = env_attr(:invalidator)
      return false unless invalidator_data

      invalidator_class = Invalidator.find(invalidator_data.keys.first)
      if invalidator_class
        @_invalidator = invalidator_class.new invalidator_data.values.first
        return @_invalidator
      else
        raise ArgumentError, "no invalidators registered for key :#{invalidator_data.keys.first}"
      end
    end

    private

      class Remote < Struct.new(:host, :root, :url,
                                :dir_permissions, :file_permissions, :rsync_flags)
      end

      def default_remote
        {
          host: host,
          root: root,
          url:  url,
          dir_permissions:  dir_permissions,
          file_permissions: file_permissions,
          rsync_flags:      rsync_flags
        }
      end

      def add_env_path_modifications_to_env_data environments
        Hash[*environments.map do |env,data|
          data = data.dup
          unless env == :production
            data[:root] = File.join( data[:root], branch, '/' )
            data[:url]  = File.join( data[:url],  branch )
          end

          [env, data]
        end.flatten]
      end

      def env_for_branch
        if branch == production_branch
          return :production
        elsif raw_data.fetch(:environments, {}).has_key?( branch.to_sym )
          return branch.to_sym
        else
          return raw_data.fetch(:default_env, 'staging').to_sym
        end
      end

      def normalize_data data
        if data.respond_to?(:to_hash)
          symbolize_keys data.to_hash
        else
          symbolize_keys read_config_file(data)
        end
      end

      def read_config_file path
        if File.exists?(path)
          YAML.load ERB.new(IO.read(path)).result
        else
          raise "config file not found at #{path}"
        end
      end

      def symbolize_keys(hash)
        hash.inject({}) do |out, (key, value)|
          k = case key
              when String then key.to_sym
              else key
              end
          v = case value
              when Hash then symbolize_keys(value)
              else value
              end
          out[k] = v
          out
        end
      end

  end
end
